import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;

import org.opencv.core.Core;
import org.opencv.core.Mat;
import org.opencv.core.MatOfRect;
import org.opencv.core.Point;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.highgui.Highgui;
import org.opencv.highgui.VideoCapture;
import org.opencv.imgproc.Imgproc;
import org.opencv.objdetect.CascadeClassifier;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
//import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;
//import javax.swing.ImageIcon;
//import javax.swing.JFrame;
//import javax.swing.JLabel;




//import org.opencv.core.Core;
import org.opencv.core.CvType;
//import org.opencv.core.Mat;
import org.opencv.core.MatOfByte;
import org.opencv.core.MatOfPoint;


public class FaceTracking {

	protected String command = ""; // The command
	protected String previousCommand = "NULL"; // Previous command
	protected double x; // Cartesian coordinates of the red spoted item
	protected double y;
	protected double area;
	protected boolean graphic; // If it's false, we wont't have any graphic frame (and save memory so far)
	protected boolean trackOk; // Enable/Disable tracking to go out the routine loop
	protected int success; // Number of tracking consecutive command success (we need N similar tracking command to get the command actually applied)

	public FaceTracking() {
		// Default constructor
		command = ""; // No command for the begining
		previousCommand = "";
		x = 300; // Point corresponding approx. to the center of the image...   
		y = 160; // and...
		area = 20000; // ...the item is enough close to the webcam
		// so, the Tracking will return a NULL command (red spot image perfectly centered)
		graphic = false;
		trackOk = false; // Caution ! You must enable tracking (TrackOk) and graphics (graphic) in view to set your tracking algoritm
		success = 0; // Number of success initialized to 0 

	}

	static Mat imag = null;
	
	public void track() {
		if(this.graphic) {
			trackWithGraphic();
		}
		else {
			trackWithoutGraphic();
		}
	}

	public void trackWithGraphic() {

		//load opencv native library
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

		String face_cascade_name = "haarcascade_frontalface_alt.xml";
		String eyes_cascade_name = "haarcascade_eye_tree_eyeglasses.xml";
		CascadeClassifier face_cascade = new CascadeClassifier();
		CascadeClassifier eyes_cascade = new CascadeClassifier();
		String window_name = "Capture - Face detection.jpg";

		JFrame jframe = new JFrame(window_name);
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		JLabel vidpanel = new JLabel();
		jframe.setContentPane(vidpanel);
		jframe.setSize(640, 480);
		jframe.setVisible(true);

		System.out.println("capture through camera "+Core.VERSION);


		//load the face xml cascade
		if(!face_cascade.load(face_cascade_name))
		{
			System.out.println("Error loading face cascade");
		}
		else
		{
			System.out.println("Success loading face cascade");
		}

		//load the eyes xml cascade
		if(!eyes_cascade.load(eyes_cascade_name))
		{
			System.out.println("Error loading eyes cascade");
		}
		else
		{
			System.out.println("Success loading eyes cascade");
		}

		//detect default camera
		VideoCapture capture = new VideoCapture(0);

		if(!capture.isOpened())
		{
			System.out.println("Did not connected to camera.");
		}
		else
		{
			System.out.println("Conected to camera: "+capture.toString());
		}

		Size sz = new Size(640, 480);
		Mat outerBox = new Mat();

		//create new Mat image
		while(this.trackOk) {
			Mat frame = new Mat();
			capture.retrieve(frame);

			if(capture.read(frame)) {

				Mat frame_gray = new Mat();
				Imgproc.resize(frame, frame, sz);
				imag = frame.clone();
				outerBox = new Mat(frame.size(), CvType.CV_8UC1);
				Imgproc.cvtColor(frame, frame_gray, Imgproc.COLOR_BGRA2GRAY);
				Imgproc.equalizeHist(frame_gray, frame_gray);


				MatOfRect faces = new MatOfRect();

				face_cascade.detectMultiScale(frame_gray, faces, 1.1, 2, 0, new Size(30,30), new Size() );


				Rect[] facesArray = faces.toArray();
				int i=0;
				while( i<facesArray.length)
				{
					Point center = new Point(facesArray[i].x + facesArray[i].width * 0.5, facesArray[i].y + facesArray[i].height * 0.5);
					Core.ellipse(imag, center, new Size(facesArray[i].width * 0.5, facesArray[i].height * 0.5), 0, 0, 360, new Scalar(255, 0, 255), 4, 8, 0);

					Mat faceROI = frame_gray.submat(facesArray[i]);
					MatOfRect eyes = new MatOfRect();

					//-- In each face, detect eyes
					eyes_cascade.detectMultiScale(faceROI, eyes, 1.1, 2, 0,new Size(30,30), new Size());            

					Rect[] eyesArray = eyes.toArray();

					System.out.println("taille tableau" + eyesArray.length);
					int j = 0;
					int size_tab = eyesArray.length;
					while (j < size_tab )
					{
						System.out.println("j = " + j);
						Point center1 = new Point(facesArray[i].x + eyesArray[j].x + eyesArray[j].width * 0.5, facesArray[i].y + eyesArray[j].y + eyesArray[j].height * 0.5);
						int radius = (int) Math.round((eyesArray[j].width + eyesArray[j].height) * 0.25);
						Core.circle(imag, center1, radius, new Scalar(255, 0, 0), 4, 8, 0);
						j++;
					}
				
					// Refreshing of the atributes
				    this.x = center.x;
				    this.y = center.y;
				    this.area = facesArray[i].width*facesArray[i].height; //not really the area of the ellipse but of the box surrounding it
				    // if we detect eyes, the robot cannot goFORWARD
				    if(size_tab > 0) {
				    	this.area = 4e6; //enormous value in order to say "we are close enough" to Augustin Bot
				    }
					corrections(this.x,this.y,this.area);
					//System.out.println(area);
					System.out.println(this.getCommand());
				    i++;
				}






			}
			ImageIcon image = new ImageIcon(Mat2bufferedImage(imag));
			vidpanel.setIcon(image);
			vidpanel.repaint();
			Highgui.imwrite(window_name, frame);
		}


		capture.release();
	}

	public void trackWithoutGraphic() {

		//load opencv native library
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);

		String face_cascade_name = "haarcascade_frontalface_alt.xml";
		String eyes_cascade_name = "haarcascade_eye_tree_eyeglasses.xml";
		CascadeClassifier face_cascade = new CascadeClassifier();
		CascadeClassifier eyes_cascade = new CascadeClassifier();
		String window_name = "Capture - Face detection.jpg";


		System.out.println("capture through camera "+Core.VERSION);


		//load the face xml cascade
		if(!face_cascade.load(face_cascade_name))
		{
			System.out.println("Error loading face cascade");
		}
		else
		{
			System.out.println("Success loading face cascade");
		}

		//load the eyes xml cascade
		if(!eyes_cascade.load(eyes_cascade_name))
		{
			System.out.println("Error loading eyes cascade");
		}
		else
		{
			System.out.println("Success loading eyes cascade");
		}

		//detect default camera
		VideoCapture capture = new VideoCapture(0);

		if(!capture.isOpened())
		{
			System.out.println("Did not connected to camera.");
		}
		else
		{
			System.out.println("Conected to camera: "+capture.toString());
		}

		Size sz = new Size(640, 480);
		Mat outerBox = new Mat();

		//create new Mat image
		while(this.trackOk) {
			Mat frame = new Mat();
			capture.retrieve(frame);

			if(capture.read(frame)) {

				Mat frame_gray = new Mat();
				Imgproc.resize(frame, frame, sz);
				imag = frame.clone();
				outerBox = new Mat(frame.size(), CvType.CV_8UC1);
				Imgproc.cvtColor(frame, frame_gray, Imgproc.COLOR_BGRA2GRAY);
				Imgproc.equalizeHist(frame_gray, frame_gray);


				MatOfRect faces = new MatOfRect();

				face_cascade.detectMultiScale(frame_gray, faces, 1.1, 2, 0, new Size(30,30), new Size() );


				Rect[] facesArray = faces.toArray();
				int i=0;
				while( i<facesArray.length)
				{
					Point center = new Point(facesArray[i].x + facesArray[i].width * 0.5, facesArray[i].y + facesArray[i].height * 0.5);
					Core.ellipse(imag, center, new Size(facesArray[i].width * 0.5, facesArray[i].height * 0.5), 0, 0, 360, new Scalar(255, 0, 255), 4, 8, 0);

					Mat faceROI = frame_gray.submat(facesArray[i]);
					MatOfRect eyes = new MatOfRect();

					//-- In each face, detect eyes
					eyes_cascade.detectMultiScale(faceROI, eyes, 1.1, 2, 0,new Size(30,30), new Size());            

					Rect[] eyesArray = eyes.toArray();

					System.out.println("taille tableau" + eyesArray.length);
					int j = 0;
					int size_tab = eyesArray.length;
					while (j < size_tab )
					{
						System.out.println("j = " + j);
						Point center1 = new Point(facesArray[i].x + eyesArray[j].x + eyesArray[j].width * 0.5, facesArray[i].y + eyesArray[j].y + eyesArray[j].height * 0.5);
						int radius = (int) Math.round((eyesArray[j].width + eyesArray[j].height) * 0.25);
						Core.circle(imag, center1, radius, new Scalar(255, 0, 0), 4, 8, 0);
						j++;
					}
				
					// Refreshing of the atributes
				    this.x = center.x;
				    this.y = center.y;
				    this.area = facesArray[i].width*facesArray[i].height; //not really the area of the ellipse but of the box surrounding it
				    // if we detect eyes, the robot cannot goFORWARD
				    if(size_tab > 0) {
				    	this.area = 4e6; //enormous value in order to say "we are close enough" to Augustin Bot
				    }
					corrections(this.x,this.y,this.area);
					//System.out.println(area);
					System.out.println(this.getCommand());
				    i++;
				}






			}
			
		}


		//capture.release();
	}




	public static BufferedImage Mat2bufferedImage(Mat image) {
		MatOfByte bytemat = new MatOfByte();
		Highgui.imencode(".jpg", image, bytemat);
		byte[] bytes = bytemat.toArray();
		InputStream in = new ByteArrayInputStream(bytes);
		BufferedImage img = null;
		try {
			img = ImageIO.read(in);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return img;
	}

	public static ArrayList<Rect> detection_contours(Mat outmat) {
		Mat v = new Mat();
		Mat vv = outmat.clone();
		List<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		Imgproc.findContours(vv, contours, v, Imgproc.RETR_LIST,
				Imgproc.CHAIN_APPROX_SIMPLE);

		double maxArea = 100;
		int maxAreaIdx = -1;
		Rect r = null;
		ArrayList<Rect> rect_array = new ArrayList<Rect>();

		for (int idx = 0; idx < contours.size(); idx++) { Mat contour = contours.get(idx); double contourarea = Imgproc.contourArea(contour); if (contourarea > maxArea) {
			// maxArea = contourarea;
			maxAreaIdx = idx;
			r = Imgproc.boundingRect(contours.get(maxAreaIdx));
			rect_array.add(r);
			Imgproc.drawContours(imag, contours, maxAreaIdx, new Scalar(0,0, 255));
		}

		}

		v.release();

		return rect_array;
	}


	//Those three following methods are returning booleans to mention if 
	//correction have been used of not

	public boolean correction_rotation(double x) {
		// First, we will correct the angular position of the robot in the (x,y) plan
		// For this, we have two threesold in the x axis, x needs to be on this range
		double x_low = 250.0;
		double x_high = 440.0;
		boolean b = false;
		if(x_low > x) {
			this.command = "toRIGHT";
			b = true;
		}
		else if(x_high < x) {
			this.command = "toLEFT";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public boolean correction_translation(double area) {
		// If the area of the circular object is too little then the item is too far
		// -> The robots needs to approach the object
		double proximity_area = 20000.0;
		boolean b = false;
		if(proximity_area>area) {
			this.command = "toFORWARD";
			b = true;
		}
		else {
			this.command = "NULL";
		}
		return b;
	}

	public boolean correction_inclinaison(double y) {
		// idem than correction_translation() but here for the y axis
		double y_low = 250.0;
		double y_high = 350.0;
		boolean b = false;
		if(y_low > y) {
			this.command = "toDOWN";

			b = true;
		}
		else if(y_high < y) {
			this.command = "toTOP";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public void corrections (double x, double y, double d) {
		// We are going to correct only one parameter per iteration
		// 1 : x axis
		// 2 : y axis
		// 3 : z axis (area, approach of the object)

		boolean a = correction_rotation(x);
		if(!a) {
			boolean b = correction_inclinaison(y);
			if(!b){
				correction_translation(d);
			}
		}
	}

	public String getCommand() {
		// To get the command is compulsory to determine if the command is feasible 
		// (see next method)
		String s;
		if(this.checkCommand()) {
			s = this.command; 
		}
		else {
			s = "NULL";
		}
		return s;
	}

	public boolean checkCommand() {
		// If 5 attempted commands are the same, then your tracking suceed this test
		// This test is coded to avoid noise effect (tracking errors) as a low pass filter
		boolean b = false;
		if(this.previousCommand.equals(this.command)) {
			this.success++;
		}
		else {
			this.success = 0;
		}
		if(this.success>4){
			b = true;
		}
		else {
			b = false;
		}

		System.out.println("command precedente" + this.previousCommand);
		this.previousCommand = this.command;
		return b;
	}

	public void trackingFollowed() {
		if(this.command.equals("toTOP")){
			//this.trackOk = false;
			//System.out.println("Sortie du process");
		}
	}

	// Two methods for external command of tracking
	public void enableTrack() {
		this.trackOk = true;
	}
	public void disableTrack() {
		this.trackOk = false;
	}

	// Two methods to enable/disable graphics
	public void enableGraphic() {
		this.graphic = true;
	}
	public void disableGraphic () {
		this.graphic = false;
	}

}



