
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;

import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfByte;
import org.opencv.core.MatOfPoint;
import org.opencv.core.Rect;
import org.opencv.core.Scalar;
import org.opencv.core.Size;
import org.opencv.highgui.Highgui;
import org.opencv.highgui.VideoCapture;
import org.opencv.imgproc.Imgproc;




public class HumanMotion {

	protected double x_mean;
	protected double y_mean;
	protected double area_mean;
	protected double area_sum;
	protected int iterations;
	protected int success;
	protected String command;
	protected String previousCommand;

	static {
		System.loadLibrary(Core.NATIVE_LIBRARY_NAME);
	}

	static Mat imag = null;

	public HumanMotion() {
		x_mean = 0.0;
		y_mean = 0.0;
		area_mean = 0.0;
		area_sum = 0.0;
		iterations = 0;
		success = 0;
		command = "";
		previousCommand = "";
	}

	public void track() {

		JFrame jframe = new JFrame("Augustin Human Motion Detector");
		jframe.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		JLabel vidpanel = new JLabel();
		jframe.setContentPane(vidpanel);
		jframe.setSize(640, 480);
		jframe.setVisible(true);
		
		
		Mat frame = new Mat();
		Mat outerBox = new Mat();
		Mat diff_frame = null;
		Mat tempon_frame = null;
		ArrayList<Rect> array = new ArrayList<Rect>();
		VideoCapture camera = new VideoCapture(0);
		Size sz = new Size(640, 480);
		int i = 0;

		while (true) {
			x_mean = 300.0;
			y_mean = 300.0;
			area_mean = 20000.0;
			iterations = 0;
			if (camera.read(frame)) {
				Imgproc.resize(frame, frame, sz);
				imag = frame.clone();
				outerBox = new Mat(frame.size(), CvType.CV_8UC1);
				Imgproc.cvtColor(frame, outerBox, Imgproc.COLOR_BGR2GRAY);
				Imgproc.GaussianBlur(outerBox, outerBox, new Size(3, 3), 0);
				
				
				
				if (i == 0) {
					jframe.setSize(frame.width(), frame.height());
					diff_frame = new Mat(outerBox.size(), CvType.CV_8UC1);
					tempon_frame = new Mat(outerBox.size(), CvType.CV_8UC1);
					diff_frame = outerBox.clone();
				}

				if (i == 1) {
					Core.subtract(outerBox, tempon_frame, diff_frame);
					Imgproc.adaptiveThreshold(diff_frame, diff_frame, 255,
							Imgproc.ADAPTIVE_THRESH_MEAN_C,
							Imgproc.THRESH_BINARY_INV, 5, 2);
					array = detection_contours(diff_frame);
					if (array.size() > 0) {

						Iterator<Rect> it2 = array.iterator();
						boolean more_than_one = false;
			
						while (it2.hasNext()) {
							more_than_one = true;
							Rect obj = it2.next();
							Core.rectangle(imag, obj.br(), obj.tl(),
									new Scalar(0, 255, 0), 1);
							double x_local = obj.x;
							double y_local = obj.y;
							double area_local = obj.area();
							double local_noise_area = 2e4;
							if(area_local < local_noise_area) {
							iterations++;
							x_mean += x_local;
							y_mean += y_local;
							area_mean += area_local;
							}
						}
						if(more_than_one && iterations != 0) {
							x_mean = x_mean/iterations;
							y_mean = y_mean/iterations;
							area_mean += area_mean/iterations;
						}

					}
		
				}

				i = 1;

				ImageIcon image = new ImageIcon(Mat2bufferedImage(imag));
				vidpanel.setIcon(image);
				vidpanel.repaint();
				tempon_frame = outerBox.clone();
				corrections(x_mean, y_mean, area_mean);
				
				

			}
			
			System.out.println(this.toString());
			System.out.println(this.getCommand());
		}
	}

	public static BufferedImage Mat2bufferedImage(Mat image) {
		MatOfByte bytemat = new MatOfByte();
		Highgui.imencode(".jpg", image, bytemat);
		byte[] bytes = bytemat.toArray();
		InputStream in = new ByteArrayInputStream(bytes);
		BufferedImage img = null;
		try {
			img = ImageIO.read(in);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return img;
	}

	public static ArrayList<Rect> detection_contours(Mat outmat) {
		Mat v = new Mat();
		Mat vv = outmat.clone();
		List<MatOfPoint> contours = new ArrayList<MatOfPoint>();
		Imgproc.findContours(vv, contours, v, Imgproc.RETR_LIST,
				Imgproc.CHAIN_APPROX_SIMPLE);

		double maxArea = 100;
		int maxAreaIdx = -1;
		Rect r = null;
		ArrayList<Rect> rect_array = new ArrayList<Rect>();

		for (int idx = 0; idx < contours.size(); idx++) { Mat contour = contours.get(idx); double contourarea = Imgproc.contourArea(contour); if (contourarea > maxArea) {
			// maxArea = contourarea;
			maxAreaIdx = idx;
			r = Imgproc.boundingRect(contours.get(maxAreaIdx));
			rect_array.add(r);
			Imgproc.drawContours(imag, contours, maxAreaIdx, new Scalar(0,0, 255));
		}

		}

		v.release();

		return rect_array;

	}



//Those three following methods are returning booleans to mention if 
//correction have been used of not

	public boolean correction_rotation(double x) {
		// First, we will correct the angular position of the robot in the (x,y) plan
		// For this, we have two threesold in the x axis, x needs to be on this range
		double x_low = 250.0;
		double x_high = 440.0;
		boolean b = false;
		if(x_low > x) {
			this.command = "toRIGHT";
			b = true;
		}
		else if(x_high < x) {
			this.command = "toLEFT";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public boolean correction_translation(double area) {
		// If the area of the circular object is too little then the item is too far
		// -> The robots needs to approach the object
		double proximity_area = 8e4;
		double noise_area = 1e4;
		boolean b = false;
		if(proximity_area>area && noise_area<area) {
			this.command = "toFORWARD";
			b = true;
		}
		else {
			this.command = "NULL";
		}
		return b;
	}

	public boolean correction_inclinaison(double y) {
		// idem than correction_translation() but here for the y axis
		double y_low = 250.0;
		double y_high = 350.0;
		boolean b = false;
		if(y_low > y) {
			this.command = "toDOWN";

			b = true;
		}
		else if(y_high < y) {
			this.command = "toTOP";
			b = true;
		}
		else {
			this.command = "NULL";
		}

		return b;
	}

	public void corrections (double x, double y, double d) {
		// We are going to correct only one parameter per iteration
		// 1 : x axis
		// 2 : y axis
		// 3 : z axis (area, approach of the object)
		
		
		boolean a = correction_rotation(x);
		if(!a) {
			boolean b = correction_inclinaison(y);
			if(!b){
				correction_translation(d);
			}
		}
	}

	public String getCommand() {
		// To get the command is compulsory to determine if the command is feasible 
		// (see next method)
		String s;
		if(this.checkCommand()) {
			s = this.command; 
			
			if (s.equals("toFORWARD")){
				try {
				    Thread.sleep(1000);                 //100 milliseconds is one second.
				} catch(InterruptedException ex) {
				    Thread.currentThread().interrupt();
				}
			}
			else if(!s.equals("NULL")){
				try {
				    Thread.sleep(500);                 //100 milliseconds is one second.
				} catch(InterruptedException ex) {
				    Thread.currentThread().interrupt();
				}
			}
		}
		else {
			s = "NULL";
		}
		return s;
	}

	public boolean checkCommand() {
		// If 5 attempted commands are the same, then your tracking suceed this test
		// This test is coded to avoid noise effect (tracking errors) as a low pass filter
		boolean b = false;
		double areaThreesold = 1e5;
		
		if(this.previousCommand.equals(this.command)) {
			this.success++;
			this.area_sum += this.area_mean;
		}
		else {
			this.success = 0;
		}
		if(this.success>4){
			if(this.area_sum > areaThreesold || !this.command.equals("toFORWARD")){
				b = true;
			}
			this.area_sum = 0.0;
		}
		else {
			b = false;
		}
		
		this.previousCommand = this.command;
		return b;
	}
	
	public void trackingFollowed() {
		if(this.command.equals("toTOP")){
			//this.trackOk = false;
			//System.out.println("Sortie du process");
		}
	}

	@Override
	public String toString() {
		return "HumanMotion [x_mean=" + x_mean + ", y_mean=" + y_mean
				+ ", area_mean=" + area_mean + ", area_sum=" + area_sum
				+ ", iterations=" + iterations + ", success=" + success
				+ ", command=" + command + ", previousCommand="
				+ previousCommand + "]";
	}
}


